#include "GLWindow.h"
#include <iostream>
#include <ngl/Light.h>
#include <ngl/Transformation.h>
#include <ngl/NGLInit.h>
#include <ngl/Obj.h>
#include <ngl/VBOPrimitives.h>
#include <ngl/ShaderManager.h>
#include <ngl/Util.h>
#include <ngl/Util.h>

const static float s_shipUpdate=0.2;
//----------------------------------------------------------------------------------------------------------------------
GLWindow::GLWindow(
                   QWidget *_parent
                  ) :
                    QGLWidget(_parent)
{
  // set this widget to have the initial keyboard focus
  setFocus();
  // re-size the widget to that of the parent (in this case the GLFrame passed in on construction)
  this->resize(_parent->size());
  // now we start a timer and assign it to the m_updateShipTimer variable
	// this will be triggered at an interval of every 5ms
	// to stop this timer we call killTimer(m_animationTimer) we can have
	// multiple timers but each must have it's own ID
	m_updateShipTimer=startTimer(2);
	m_redrawTimer=startTimer(30);
	// now clear the key set
	m_keysPressed.clear();


}

//----------------------------------------------------------------------------------------------------------------------
GLWindow::~GLWindow()
{
	// call the static NGLQuit method to close down
	// NGL and tidy up any shaders textures etc
	// first get an instance of the NGLInit singleton
	ngl::NGLInit *ngl = ngl::NGLInit::instance();
	// then call the quit method
	ngl->NGLQuit();
}

//----------------------------------------------------------------------------------------------------------------------
// This virtual function is called once before the first call to paintGL() or resizeGL(),
//and then once whenever the widget has been assigned a new QGLContext.
// This function should set up any required OpenGL context rendering flags, defining display lists, etc.
//----------------------------------------------------------------------------------------------------------------------
void GLWindow::initializeGL()
{
	// basically we do our main OpenGL setup here
	// set the clear colour in this case to black
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	// enable depth testing for drawing
	glEnable(GL_DEPTH_TEST);

	// before we can start any advanced OpenGL under windows or linux we need to
	// load in the extensions using GLEW mac will ignore this so it works for all
	// platforms, to do this we use the NGLInit class
	ngl::NGLInit *Init = ngl::NGLInit::instance();
	#ifdef WIN32
		glewInit();
	#endif
	Init->initGlew();
	// Now we will create a basic Camera from the graphics library
	// This is a static camera so it only needs to be set once
	// First create Values for the camera position
	ngl::Vector From(0,0,-80);
	ngl::Vector To(0,0,0);
	ngl::Vector Up(0,1,0);
	// now we instantiate our camera
	m_cam= new ngl::Camera(From,To,Up,ngl::ORTHOGRAPHIC);
	// set the shape using FOV 45 Aspect Ratio based on Width and Height
	// The final two are near and far clipping planes of 0.5 and 150
	m_cam->setShape(45,(float)1024/768,0.5,150,ngl::PERSPECTIVE);
	// now to load the shader and set the values
	// grab an instance of shader manager
	ngl::ShaderManager *shader=ngl::ShaderManager::instance();
	// load a frag and vert shaders
	shader->loadShader("Blinn","shaders/Vertex.vs","shaders/Fragment.fs");
	// set this as the active shader
	shader->useShader("Blinn");
	// now pass the modelView and projection values to the shader
	shader->setShaderParamFromMatrix("Blinn","ViewMatrix",m_cam->getModelView());
	shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());
  // create our spaceship
  m_ship= new SpaceShip(ngl::Vector(0,0,0,1),"models/SpaceShip.obj");
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget has been resized.
// The new size is passed in width and height.
//----------------------------------------------------------------------------------------------------------------------
void GLWindow::resizeGL(
                        int _w,
                        int _h
                       )
{
  glViewport(0,0,_w,_h);
  m_cam->setShape(45,(float)_w/_h,0.5,150,ngl::PERSPECTIVE);
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  shader->setShaderParamFromMatrix("Blinn","projectionMatrix",m_cam->getProjection());
}

//----------------------------------------------------------------------------------------------------------------------
//This virtual function is called whenever the widget needs to be painted.
// this is our main drawing routine
//----------------------------------------------------------------------------------------------------------------------
void GLWindow::paintGL()
{
  // clear the screen and depth buffer
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
 // build our transform stack
  ngl::Transformation trans;
  // set this in the TX stack
  m_transformStack.setGlobal(trans);
  // now load these values to the shader
  ngl::ShaderManager *shader=ngl::ShaderManager::instance();
  m_transformStack.loadGlobalAndCurrentMatrixToShader("Blinn","ModelMatrix");
  shader->setShaderParamFromMatrix("Blinn","ModelMatrix",m_transformStack.getCurrAndGlobal().getMatrix());
	m_ship->draw();

}


//----------------------------------------------------------------------------------------------------------------------
// this overloaded method is called every time the timers trigger
//----------------------------------------------------------------------------------------------------------------------
void GLWindow::timerEvent(
                          QTimerEvent *_event
                         )
{
	// the usual process is to check the event timerID and compare it to
	// any timers we have started with startTimer
	if (_event->timerId() == m_updateShipTimer)
	{
		moveShip();
	}
	if (_event->timerId() == m_redrawTimer)
	{
		updateGL();
	}

}

void GLWindow::moveShip()
{
	/// first we reset the movement values
	float xDirection=0.0;
	float yDirection=0.0;
	float rotation=0.0;
	// now we loop for each of the pressed keys in the the set
	// and see which ones have been pressed. If they have been pressed
	// we set the movement value to be an incremental value
	foreach(Qt::Key key, m_keysPressed)
	{
		switch (key)
		{
			case Qt::Key_Left :  { xDirection=s_shipUpdate; break;}
			case Qt::Key_Right : { xDirection=-s_shipUpdate; break;}
			case Qt::Key_Up :		 { yDirection=s_shipUpdate; break;}
			case Qt::Key_Down :  { yDirection=-s_shipUpdate; break;}
			case Qt::Key_R :     { rotation=1.0; break;}
			default : break;
		}
	}
	// if the set is non zero size we can update the ship movement
	// then tell openGL to re-draw
	if(m_keysPressed.size() !=0)
	{
		m_ship->move(xDirection,yDirection);
		m_ship->rotate(rotation);
	}
}
//----------------------------------------------------------------------------------------------------------------------
// in this case this method is called from the MainWindow class and passed
// the key event from that, if we didn't need to process the events in main
// window we could just overload this as the keyPressEvent and only process here.
//----------------------------------------------------------------------------------------------------------------------
void GLWindow::processKeyDown(
															 QKeyEvent *_event
															)
{
	// add to our keypress set the values of any keys pressed
	m_keysPressed += (Qt::Key)_event->key();

}


//----------------------------------------------------------------------------------------------------------------------
// in this case this method is called from the MainWindow class and passed
// the key event from that, if we didn't need to process the events in main
// window we could just overload this as the keyPressEvent and only process here.
//----------------------------------------------------------------------------------------------------------------------
void GLWindow::processKeyUp(
															 QKeyEvent *_event
															)
{
	// remove from our key set any keys that have been released
	m_keysPressed -= (Qt::Key)_event->key();
}

